
/******************************************************************************
* MODULE     : category.gen.h
* DESCRIPTION: categorial properties of domains
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef CATEGORY_H
#define CATEGORY_H
#include "Arithmetic/integer.hh"
#include "Arithmetic/rational.hh"
#include "Arithmetic/fp_real.hh"

/******************************************************************************
* Utility templates for properties classes
******************************************************************************/

// Allows compile time dispatch
template<typename T, T x> struct value_to_type {
  static const T value= x; };
struct bool_to_type_any {};
template<bool b> struct bool_to_type : bool_to_type_any {
  static const bool value= b; };
template<bool b> struct value_to_type<bool,b> : bool_to_type<b> {};
typedef bool_to_type<false> false_type;
typedef bool_to_type<true> true_type;

// Base class undefined properties
struct undefined {
  static const bool defined = false;
  typedef bool_to_type<false> defined_as_type;
};

// Base for properties with compile-time dispatch and partial specialization.
template<typename T, T x> struct define_value {
  static const bool defined = true;
  typedef bool_to_type<true> defined_as_type;
  static const T value = x;
  typedef value_to_type<T,x> value_as_type;
  inline operator T () { return x; } };

template<int n>  struct define_int  : define_value<int, n>  {};
template<bool b> struct define_bool : define_value<bool, b> {};

/******************************************************************************
* A ring is divisible if we may divide by non zero integers
******************************************************************************/

template<class T> struct has_divisible : define_bool<false> {};

template<> struct has_divisible<integer>  : define_bool<true> {};
template<> struct has_divisible<rational> : define_bool<true> {};
template<> struct has_divisible<float>    : define_bool<true> {};
template<> struct has_divisible<double>   : define_bool<true> {};
template<> struct has_divisible<fp_real>  : define_bool<true> {};
// template<> struct has_divisible<safe_real>  : define_bool<true> {};

/******************************************************************************
* A ring has special functions if exp and log have been implemented
******************************************************************************/

template<class T> struct has_special_functions : define_bool<false> {};

// template<> struct has_special_functions<float> : define_bool<true> {}; 
// template<> struct has_special_functions<double> : define_bool<true> {}; 

/******************************************************************************
* Thresholds for Karatsuba multiplication: must be powers of two
******************************************************************************/

template<class T> struct Karatsuba_threshold : undefined {};

template<> struct Karatsuba_threshold<float>     : define_int<16> {};
template<> struct Karatsuba_threshold<double>    : define_int<16> {};
template<> struct Karatsuba_threshold<integer>   : define_int<16> {};
template<> struct Karatsuba_threshold<rational>  : define_int<16> {};
template<> struct Karatsuba_threshold<fp_real>   : define_int<2> {};
// template<> struct Karatsuba_threshold<safe_real> : define_int<2> {};

/******************************************************************************
* Thresholds for FFT multiplication
******************************************************************************/

template<class T> inline int FFT_threshold () { return T::FFT_threshold; }

template<> inline int FFT_threshold<float>     () { return 2400; }
template<> inline int FFT_threshold<double>    () { return 2400; }
template<> inline int FFT_threshold<integer>   () { return 1000; }
template<> inline int FFT_threshold<rational>  () { return 1000; }
template<> inline int FFT_threshold<fp_real>   () { return 1000; }
// template<> inline int FFT_threshold<safe_real> () { return 1000; }

/******************************************************************************
* Characteristic of modular integers 
******************************************************************************/

template<class T> struct characteristic : undefined {};

// No built-in types have the property "characteristic"

#endif // defined CATEGORY_H
