
/******************************************************************************
* MODULE     : analytic.gen.cc
* DESCRIPTION: Analytic functions
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "tree.hh"
#include "Complex/complex.hh"
#include "Complex/series.cc"

typedef double R;
typedef complex<double> C;
C I(0,1);

#define INFINITE_RADIUS 1e10
#define SHRINK_RADIUS   0.8
#define SHRINK_FACTOR   10.0
#define MAGNIFY_FACTOR  1.4

/******************************************************************************
* Abstract analytic functions
******************************************************************************/

class analytic_function;
struct analytic_function_rep: public abstract_struct {
  int n;
  C*  series;

  inline analytic_function_rep (): n(0), series(NULL) {}
  inline virtual ~analytic_function_rep () { if (n!=0) delete[] series; }
  C* get_series (int order);

  virtual R bound_radius () = 0;
  virtual R bound_disk (R r) = 0;
  virtual void compute_series (int order) = 0;
  virtual analytic_function continuate (C z) = 0;
  virtual operator tree () = 0;

  virtual C evaluate (C z) = 0;
  virtual analytic_function derivative () = 0;
};

class analytic_function {
  ABSTRACT(analytic_function);
  analytic_function (C z);
  inline operator tree () { return (tree) (*rep); }
};
ABSTRACT_CODE(analytic_function);

analytic_function id (C z0);
analytic_function operator + (analytic_function f1, analytic_function f2);
analytic_function operator - (analytic_function f1, analytic_function f2);
analytic_function operator * (analytic_function f1, analytic_function f2);
analytic_function operator / (analytic_function f1, analytic_function f2);
analytic_function exp (analytic_function f1);
analytic_function log (analytic_function f1);
analytic_function log (analytic_function f1, C val);
analytic_function differentiate (analytic_function f1);
analytic_function integrate (analytic_function f1, C val=0);

/******************************************************************************
* Constant and coordinate functions
******************************************************************************/

class constant_function_rep: public analytic_function_rep {
  C z;

public:
  constant_function_rep (C z2): z(z2) {}
  ~constant_function_rep () {}

  R bound_radius () { return INFINITE_RADIUS; }
  R bound_disk (R r) { (void) r; return abs (z); }
  void compute_series (int order) {
    int i;
    series[0]= z;
    for (i=n; i<order; i++) series[i]= C(0);
  }
  analytic_function continuate (C dummy) {
    (void) dummy;
    return analytic_function (z);
  }
  operator tree () {
    string re= as_string (z.real());
    string im= as_string (z.imag());
    return tuple ("+", re, tuple ("*", im, "i"));
  }
  C evaluate (C dummy) { (void) dummy; return z; }
  analytic_function derivative () { return analytic_function (C(0)); }
};

analytic_function::analytic_function (C z) {
  rep= new constant_function_rep (z);
  rep->ref_count++;
}

class identity_function_rep: public analytic_function_rep {
  C z0;

public:
  identity_function_rep (C z0b): z0(z0b) {}
  ~identity_function_rep () {}

  R bound_radius () { return INFINITE_RADIUS; }
  R bound_disk (R r) { return abs (z0)+ r; }
  void compute_series (int order) {
    int i;
    for (i=n; i<order; i++) series[i]= C(0);
    series[0]= z0;
    if (order>1) series[1]= C(1);
  }
  analytic_function continuate (C z) { return id (z0+z); }
  operator tree () {
    return tuple ("z", as_string (z0.real()), as_string (z0.imag()));
  }
  C evaluate (C z) { return z0+ z; }
  analytic_function derivative () { return analytic_function (C(1)); }
};

analytic_function
id (C z0) {
  return new identity_function_rep (z0);
}

/******************************************************************************
* Arithmetic functions
******************************************************************************/

class plus_function_rep: public analytic_function_rep {
  analytic_function f1;
  analytic_function f2;

public:
  plus_function_rep (analytic_function f1b, analytic_function f2b):
    f1 (f1b), f2 (f2b) {}
  ~plus_function_rep () {}

  R bound_radius () { return f1->bound_radius () <? f2->bound_radius (); }
  R bound_disk (R r) { return f1->bound_disk (r)+ f2->bound_disk (r); }
  void compute_series (int order) {
    int i;
    C* s1= f1->get_series (order);
    C* s2= f2->get_series (order);
    for (i=n; i<order; i++) series[i]= s1[i]+ s2[i];
  }
  analytic_function continuate (C z) {
    return f1->continuate (z)+ f2->continuate (z);
  }
  operator tree () { return tuple ("+", (tree) f1, (tree) f2); }
  C evaluate (C z) { return f1->evaluate (z)+ f2->evaluate (z); }
  analytic_function derivative () { return f1->derivative()+ f2->derivative(); }
};

analytic_function
operator + (analytic_function f1, analytic_function f2) {
  return new plus_function_rep (f1, f2);
}

class minus_function_rep: public analytic_function_rep {
  analytic_function f1;
  analytic_function f2;

public:
  minus_function_rep (analytic_function f1b, analytic_function f2b):
    f1 (f1b), f2 (f2b) {}
  ~minus_function_rep () {}

  R bound_radius () { return f1->bound_radius () <? f2->bound_radius (); }
  R bound_disk (R r) { return f1->bound_disk (r)+ f2->bound_disk (r); }
  void compute_series (int order) {
    int i;
    C* s1= f1->get_series (order);
    C* s2= f2->get_series (order);
    for (i=n; i<order; i++) series[i]= s1[i]- s2[i];
  }
  analytic_function continuate (C z) {
    return f1->continuate (z)- f2->continuate (z);
  }
  operator tree () { return tuple ("-", (tree) f1, (tree) f2); }
  C evaluate (C z) { return f1->evaluate (z)- f2->evaluate (z); }
  analytic_function derivative () { return f1->derivative()- f2->derivative(); }
};

analytic_function
operator - (analytic_function f1, analytic_function f2) {
  return new minus_function_rep (f1, f2);
}

class times_function_rep: public analytic_function_rep {
  analytic_function f1;
  analytic_function f2;

public:
  times_function_rep (analytic_function f1b, analytic_function f2b):
    f1 (f1b), f2 (f2b) {}
  ~times_function_rep () {}

  R bound_radius () { return f1->bound_radius () <? f2->bound_radius (); }
  R bound_disk (R r) { return f1->bound_disk (r)* f2->bound_disk (r); }
  void compute_series (int order) {
    int i, j;
    C* s1= f1->get_series (order);
    C* s2= f2->get_series (order);
    for (i=n; i<order; i++) {
      series[i]= C(0);
      for (j=0; j<=i; j++)
	series[i] += s1[j]*s2[i-j];
    }
  }
  analytic_function continuate (C z) {
    return f1->continuate (z)* f2->continuate (z);
  }
  operator tree () { return tuple ("*", (tree) f1, (tree) f2); }
  C evaluate (C z) { return f1->evaluate (z)* f2->evaluate (z); }
  analytic_function derivative () {
    return f1->derivative()* f2+ f1*f2->derivative(); }
};

analytic_function
operator * (analytic_function f1, analytic_function f2) {
  return new times_function_rep (f1, f2);
}

class over_function_rep: public analytic_function_rep {
  analytic_function f1;
  analytic_function f2;
  analytic_function df2;
  R af2;
  R r;

public:
  over_function_rep (analytic_function f1b, analytic_function f2b):
    f1 (f1b), f2 (f2b), df2 (f2->derivative ())
  {
    af2 = abs (f2->evaluate (0.0));
    R r0= (f1->bound_radius() <? f2->bound_radius ())* SHRINK_RADIUS; r= r0;
    while ((df2->bound_disk (r)*r) > af2) r /= SHRINK_FACTOR;
    while ((r*MAGNIFY_FACTOR<r0) &&
	   ((df2->bound_disk (r*MAGNIFY_FACTOR)*r*MAGNIFY_FACTOR) <= af2))
      r *= MAGNIFY_FACTOR;
  }
  ~over_function_rep () {}

  R bound_radius () { return r; }
  R bound_disk (R r) {
    R variation= df2->bound_disk (r)* r;
    return f1->bound_disk (r)/ (af2- variation);
  }
  void compute_series (int order) {
    C* s1= f1->get_series (order);
    C* s2= f2->get_series (order);
    divide (series, s1, s2, order);
  }
  analytic_function continuate (C z) {
    return f1->continuate (z) / f2->continuate (z);
  }
  operator tree () { return tuple ("/", (tree) f1, (tree) f2); }
  C evaluate (C z) { return f1->evaluate (z) / f2->evaluate (z); }
  analytic_function derivative () {
    return (f1->derivative()*f2- f1*df2)/(f2*f2); }
};

analytic_function
operator / (analytic_function f1, analytic_function f2) {
  return new over_function_rep (f1, f2);
}

/******************************************************************************
* Exponentiation and logarithm
******************************************************************************/

static R LOG_INFINITE_RADIUS= log (INFINITE_RADIUS);

class exp_function_rep: public analytic_function_rep {
  analytic_function f1;
  R r;

public:
  exp_function_rep (analytic_function f1b): f1 (f1b) {
    if (abs (f1->evaluate (C(0))) > LOG_INFINITE_RADIUS) r=1.0/INFINITE_RADIUS;
    else {
      R r0= f1->bound_radius ()*SHRINK_RADIUS; r=r0;
      while (f1->bound_disk (r) >= LOG_INFINITE_RADIUS) r /= SHRINK_FACTOR;
      while ((r*MAGNIFY_FACTOR<r0) &&
	     (f1->bound_disk (r*MAGNIFY_FACTOR) < LOG_INFINITE_RADIUS))
	r *= MAGNIFY_FACTOR;
    }
  }
  ~exp_function_rep () {}

  R bound_radius () { return r; }
  R bound_disk (R r) { return exp (f1->bound_disk (r)); }
  void compute_series (int order) {
    C* s1= f1->get_series (order);
    exponentiate (series, s1, order);
  }
  analytic_function continuate (C z) {
    return exp (f1->continuate (z));
  }
  operator tree () { return tuple ("exp", (tree) f1); }
  C evaluate (C z) { return exp (f1->evaluate (z)); }
  analytic_function derivative () { return f1->derivative()* exp (f1); }
};

analytic_function
exp (analytic_function f1) {
  return new exp_function_rep (f1);
}

class log_function_rep: public analytic_function_rep {
  analytic_function f1;
  analytic_function df1;
  C lf1;
  R af1;
  R r;

public:
  log_function_rep (analytic_function f1b, C lf1b):
    f1 (f1b), df1 (f1->derivative ()), lf1 (lf1b)
  {
    af1 = abs (f1->evaluate (0.0));
    R r0= f1->bound_radius()* SHRINK_RADIUS; r= r0;
    while ((df1->bound_disk (r)*r) > af1) r /= SHRINK_FACTOR;
    while ((r*MAGNIFY_FACTOR<r0) &&
	   ((df1->bound_disk (r*MAGNIFY_FACTOR)*r*MAGNIFY_FACTOR) <= af1))
      r *= MAGNIFY_FACTOR;
  }
  ~log_function_rep () {}

  R bound_radius () { return r; }
  R bound_disk (R r) {
    R delta= df1->bound_disk (r)*r;
    return abs (lf1)- log (1- (delta/af1));
  }
  void compute_series (int order) {
    C* s1= f1->get_series (order);
    logarithm (series, s1, order);
  }
  analytic_function continuate (C z) {
    analytic_function f2= f1->continuate (z);
    C f1z= f1->evaluate (0.0);
    C f2z= f2->evaluate (0.0);
    return log (f2, lf1+ log (f2z/f1z));
  }
  operator tree () { return tuple ("log", (tree) f1); }
  C evaluate (C z) {
    C f1z= f1->evaluate (0.0);
    C f2z= f1->evaluate (z);
    return lf1+ log (f2z/f1z);
  }
  analytic_function derivative () { return df1/f1; }
};

analytic_function
log (analytic_function f1) {
  return new log_function_rep (f1, log (f1->evaluate (0.0)));
}

analytic_function
log (analytic_function f1, C val) {
  return new log_function_rep (f1, val);
}

/******************************************************************************
* Differentiation and integration
******************************************************************************/

class int_function_rep: public analytic_function_rep {
  analytic_function f1;
  C val;
  R r;

public:
  int_function_rep (analytic_function f1b, C valb):
    f1 (f1b), val (valb) { r= f1->bound_radius(); }
  ~int_function_rep () {}

  R bound_radius () { return r; }
  R bound_disk (R r) { return f1->bound_disk (r)*r; }
  void compute_series (int order) {
    int i;
    C* s1= f1->get_series (order-1);
    series[0]=val;
    for (i= (n>?1); i<order; i++) series[i]= s1[i-1]/C(i);
  }
  analytic_function continuate (C z) {
    return integrate (f1->continuate (z), evaluate (z));
  }
  operator tree () { return tuple ("int", (tree) f1); }
  C evaluate (C z) {
    R shrink = 1.0 / ((abs(z)/r) >? 0.1);
    int steps= (int) (40.0/log (shrink));
    if (steps>=500) {
      cout << "abs(z)= " << abs(z) << "\n";
      cout << "r     = " << r << "\n";
      cout << "steps = " << steps << "\n";
      steps= 500;
    }
    (void) get_series (steps);
    return eval_series (series, steps, z);
  }
  analytic_function derivative () { return f1; }
};

analytic_function
integrate (analytic_function f1, C val) {
  return new int_function_rep (f1, val);
}

/******************************************************************************
* Routines for analytic functions
******************************************************************************/

C*
analytic_function_rep::get_series (int order) {
  if (order>n) {
    int i;
    C* series2= new C[order];
    for (i=0; i<n; i++) series2[i]= series[i];
    if (n!=0) delete[] series;
    series= series2;
    compute_series (order);
    n= order;
  }
  return series;
}

analytic_function
test_function () {
  /*
  analytic_function z= id (C(2,0));
  return log (z-log (z));
  */

  analytic_function z= id (C(2,0));
  analytic_function log_z= log (z);
  analytic_function lambda (C(1.0));
  analytic_function alpha (C(0.0,3.14));
  analytic_function minus_alpha (C(0.0,-3.14));
  return z*(lambda+ exp(alpha*log_z)+ exp(minus_alpha*log_z));

  /*
  analytic_function f= integrate (exp(z)/z);
  return f;
  */
}

C
evaluate_straight (analytic_function f, C z) {
  // cout << "Eval straight " << ((tree) f) << " in " << z << "\n";
  R r = f->bound_radius ();
  if (abs (z) < r*SHRINK_RADIUS) return f->evaluate (z);
  if (r < (1/INFINITE_RADIUS)) return f->evaluate (0.0);
  C zz= (SHRINK_RADIUS*r/abs(z))*z;
  return evaluate_straight (f->continuate (zz), z-zz);
}

analytic_function
continuate_straight (analytic_function f, C z) {
  R r = f->bound_radius ();
  if (abs (z) < r*SHRINK_RADIUS) return f->continuate (z);
  if (r < (1/INFINITE_RADIUS)) return f;
  C zz= (SHRINK_RADIUS*r/abs(z))*z;
  return continuate_straight (f->continuate (zz), z-zz);
}

