
/******************************************************************************
* MODULE     : compose.gen.cc
* DESCRIPTION: truncated composition of relaxed power series
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef TRUNCATED_COMPOSE_CC
#define TRUNCATED_COMPOSE_CC

/******************************************************************************
* Composition with powers of z
******************************************************************************/

template<class C> struct Mahler_series_rep : public series_rep<C> {
  series<C> f;
  int p;
  Mahler_series_rep (series<C> f2, int p2):
    series_rep<C> (RELAXED), f(f2), p(p2) {}
  C next () { if ((n%p)==0) return f[n/p]; else return 0; }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    f->set_precision ((terms+p-1)/p);
  }
  operator tree () { return tuple ("o", f, tuple ("^", "z", as_string (p))); }
};

template<class C> struct progression_series_rep : public series_rep<C> {
  series<C> f;
  int p, offset;
  progression_series_rep (series<C> f2, int p2, int offset2):
    series_rep<C> (RELAXED), f(f2), p(p2), offset(offset2) {}
  C next () { return f[n*p+offset]; }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    f->set_precision (terms*p+offset);
  }
  operator tree () {
    return tuple ("progression", f, as_string (p), as_string (offset)); }
};

template<class C> static C
binary_power (C c, int i) {
  if (i==0) return 1;
  if (i==1) return c;
  C pw= binary_power (c, i>>1);
  if ((i&1)==0) return pw*pw;
  else return c*pw*pw;
}

template<class C> struct inpower_series_rep : public series_rep<C> {
  series<C> f;
  int p;
  inpower_series_rep (series<C> f2, int p2):
    series_rep<C> (RELAXED), f(f2), p(p2) {}
  C next () { return binary_power (f[n], p); }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    f->set_precision (terms);
  }
  operator tree () { return tuple ("inpower", f, as_string (p)); }
};

template<class C> series<C>
Mahler (series<C> f, int p) {
  return new Mahler_series_rep<C> (f, p);
}

template<class C> series<C>
progression (series<C> f, int p, int offset) {
  return new progression_series_rep<C> (f, p, offset);
}

template<class C> series<C>
inpower (series<C> f, int p) {
  return new inpower_series_rep<C> (f, p);
}

/******************************************************************************
* Naive composition
******************************************************************************/

template<class C> struct compose1_series_rep : public series_rep<C> {
  series<C> f, g, h;
  compose1_series_rep (series<C> f2, series<C> g2):
    series_rep<C> (RELAXED), f(f2), g(g2) {}
  C next () {
    if (n==0) return f[0];
    if (n==1) h= compose (f>>1, g) * (g>>1);
    return h[n-1];
  }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    f->set_precision (terms);
    g->set_precision (terms);
  }
  operator tree () { return tuple ("o", f, g); }
};

/******************************************************************************
* Right composition with polynomials
******************************************************************************/

template<class C> struct partial_comp_pol_series_rep : public series_rep<C> {
  series<C> f, h;
  int i, N;
  partial_comp_pol_series_rep (
    series<C> f2, int i2, int p, int q, series<C>* H);
  C next ();
  void set_precision (int terms) {
    // cout << "> partial_comp_pol_series " << terms << "\n";
    if (::N(value)<terms) value= value (0, terms);
    if (N>1) h->set_precision (terms <? N);
  }
  operator tree () { return tree ("partial-polynomial-o"); }
};

template<class C> partial_comp_pol_series_rep<C>::partial_comp_pol_series_rep (
  series<C> f2, int i2, int p, int q, series<C>* H):
    series_rep<C> (RELAXED), f (f2), i (i2), N ((p-1)*(q-1)+1)
{
  if (p==1) return;
  series<C> h1= new partial_comp_pol_series_rep<C> (f, i, p>>1, q, H);
  series<C> h2= new partial_comp_pol_series_rep<C> (f, i+(p>>1), p>>1, q, H);
  h= h1+ ((h2 * (H[p>>1]>>(p>>1))) << (p>>1));
}

template<class C> C
partial_comp_pol_series_rep<C>::next () {
  int k=n;
  if (k>0) while ((k&1)==0) k= k>>1;
  if (k==1) value= value (0, ((N <? (n<<1)) >? ::N(value)));
  if (n==N) h= series<C> ();
  if (n>=N) return 0;
  if (N==1) return f[i];
  return h[n];
}

template<class C> struct comp_pol_series_rep : public series_rep<C> {
  series<C> f, *H, h;
  int q;
  comp_pol_series_rep (series<C> f, series<C> g, int q);
  ~comp_pol_series_rep () { delete[] H; }
  C next ();
  void set_precision (int terms) {
    // cout << "> comp_pol_series " << terms << "\n";
    if (N(value)<terms) value= value (0, terms);
    h->set_precision (terms);
  }
  operator tree () { return tuple ("o", f, H[1]); }
};

template<class C> comp_pol_series_rep<C>::comp_pol_series_rep (
  series<C> f2, series<C> g, int q2):
    series_rep<C> (RELAXED), f (f2), q (q2)
{
  H= new series<C> [2];
  H[1]= g;
  h= new partial_comp_pol_series_rep<C> (f, 0, 1, q, H);
}

template<class C> C
comp_pol_series_rep<C>::next () {
  int i, k=n;
  if (k>0) while ((k&1)==0) k= k>>1;
  if (k==1) {
    if (n>1) {
      series<C>* H2= new series<C> [n+1];
      for (i=1; i<n; i=i<<1) H2[i]= H[i];
      H2[n]= H[n>>1]*H[n>>1];
      delete[] H; H=H2;
    }
    h= new partial_comp_pol_series_rep<C> (f, 0, n<<1, q, H);
    h->set_precision (N(value));
  }
  return ((series<C>) h) [n];
}

/******************************************************************************
* General composition
******************************************************************************/

// if has_divisible<C>()
template<class C> struct compose2_series_rep : public series_rep<C> {
  series<C> f, g, h;
  compose2_series_rep (series<C> f2, series<C> g2):
    series_rep<C> (RELAXED), f(f2), g(g2),
    h (new compose1_series_rep<C> (f, g)) {}
  C next ();
  void set_precision (int terms) {
    // cout << "> compose2 " << terms << "\n";
    if (N(value)<terms) value= value (0, terms);
    h->set_precision (terms);
  }
  operator tree () { return tuple ("o", f, g); }
};

template<class C> C
compose2_series_rep<C>::next () {
  if (n==0) return f[0];
  if (n==1) return f[1]*g[1];
  if (n==2) return f[2]*g[1]*g[1]+ f[1]*g[2];
  if (n==3) return (f[3]*g[1]*g[1]+ C(2)*f[2]*g[2])*g[1]+ f[1]*g[3];

  int p=0;
  while (((p+1)*(1<<(2*p))) <= n) p++;
  if ((n!=4) && ((p*(1<<(2*p-2))) <= (n-1))) return h[n];

  int nn= (p+1)*(1<<(2*p));
  int i, q= 2<<p, r= (nn+q-1)/q;
  zealous<C> g1 = g (0, q);
  zealous<C> dg1= d (g1);
  series<C>  g2 = (g- g1)>>q;

  series<C>  D  = f;
  for (i=1; i<r; i++) D= d(D);
  D= compose (D, g1)/factorial<C>(r-1);
  h= D;

  for (i=r-1; i>0; i--) {
    D= (I(C(i)*D*dg1))+ f[i-1];
    h= D+ ((h*g2)<<q);
  }
  h->set_precision (N(value));
  return h[n];
}
// end if has_divisible<C>()

/******************************************************************************
* Right composition with rational functions
******************************************************************************/

template<class C> struct partial_comp_rat_series_rep : public series_rep<C> {
  series<C> f, h;
  int i, p, N;
  partial_comp_rat_series_rep (
    series<C> f2, int i2, int p, int q, series<C>* Num, series<C>* Den);
  C next ();
  void set_precision (int terms) {
    if (::N(value)<terms) value= value (0, terms);
    if (N>1) h->set_precision (terms <? N);
  }
  operator tree () { return tree ("partial-ratynomial-o"); }
};

template<class C> partial_comp_rat_series_rep<C>::partial_comp_rat_series_rep (
  series<C> f2, int i2, int p2, int q, series<C>* Num, series<C>* Den):
    series_rep<C> (RELAXED), f (f2), i (i2), p (p2), N ((p-1)*(q-1)+1)
{
  if (p==1) return;
  series<C> h1=
    new partial_comp_rat_series_rep<C> (f, i, p>>1, q, Num, Den);
  series<C> h2=
    new partial_comp_rat_series_rep<C> (f, i+(p>>1), p>>1, q, Num, Den);
  h= (h1 * Den[p>>1]) + ((h2 * (Num[p>>1]>>(p>>1))) << (p>>1));
}

template<class C> C
partial_comp_rat_series_rep<C>::next () {
  int k=n;
  if (k>0) while ((k&1)==0) k= k>>1;
  if (k==1) value= value (0, ((N <? (n<<1)) >? ::N(value)));
  if (n==N) h= series<C> ();
  if (n>=N) return 0;
  if (N==1) return f[i];
  return h[n];
}

template<class C> struct comp_rat_series_rep : public series_rep<C> {
  series<C> f, *Num, *Den, h, Quot;
  int q;
  comp_rat_series_rep (series<C> f, series<C> num, series<C> den, int q);
  ~comp_rat_series_rep () {
    delete[] Num;
    delete[] Den;
  }
  C next ();
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    Quot->set_precision (terms);
  }
  operator tree () { return tuple ("o", f, tuple ("/", Num[1], Den[1])); }
};

template<class C> comp_rat_series_rep<C>::comp_rat_series_rep (
  series<C> f2, series<C> num, series<C> den, int q2):
    series_rep<C> (RELAXED), f (f2), q (q2)
{
  Num   = new series<C> [2];
  Den   = new series<C> [2];
  Num[1]= num;
  Den[1]= den;
  h     = new partial_comp_rat_series_rep<C> (f, 0, 1, q, Num, Den);
  Quot  = h;
}

template<class C> C
comp_rat_series_rep<C>::next () {
  int i, k=n;
  if (k>0) while ((k&1)==0) k= k>>1;
  if (k==1) {
    series<C>* Num2= new series<C> [(n<<1)+1];
    series<C>* Den2= new series<C> [(n<<1)+1];
    for (i=1; i<(n<<1); i=i<<1) {
      Num2[i]= Num[i];
      Den2[i]= Den[i];
    }
    Num2[n<<1]= Num[n]*Num[n];
    Den2[n<<1]= Den[n]*Den[n];
    if (n>1) Den2[(n<<1)-1]= Den[n]*Den[n-1];
    delete[] Num; Num=Num2;
    delete[] Den; Den=Den2;

    h= new partial_comp_rat_series_rep<C> (f, 0, n<<1, q, Num, Den);
    Quot  = h / Den[(n<<1)-1];
    Quot->set_precision (N(value));
  }
  return Quot [n];
}

/******************************************************************************
* finite characteristic
******************************************************************************/

// if characteristic<C>::defined
template<class C> struct compose3_series_rep : public series_rep<C> {
  series<C> f, g, h;
  compose3_series_rep (series<C> f2, series<C> g2):
    series_rep<C> (RELAXED), f(f2), g(g2) {}
  C next ();
  operator tree () { return tuple ("o", f, g); }
};

template<class C> C
compose3_series_rep<C>::next () {
  if (n==0) return f[0];
  if (n>1)  return h[n];

  int i, p= characteristic<C>();
  h= series<C> (C(0));
  for (i=p-1; i>=0; i--) {
    series<C> t1= progression (f, p, i);
    series<C> t2= inpower (g, p);
    series<C> t3= new compose3_series_rep<C> (t1, t2);
    series<C> t4= Mahler (t3, p);
    h= ((h*(g>>1))<<1) + t4;
  }
  h->set_precision (N(value));
  return h[n];
}
// end if characteristic<C>::defined

/******************************************************************************
* interface
******************************************************************************/

template<class C> series<C>
compose (series<C> f, zealous<C> g) {
  return new comp_pol_series_rep<C> (f, g, N(g));
}

template<class C> series<C>
compose (series<C> f, zealous<C> num, zealous<C> den) {
  return new comp_rat_series_rep<C> (f, num, den, N(num) >? N(den));
}

#define COMPOSE(HAS_CHARACTERISTIC, HAS_DIVISIBLE) \
    inline static series<C>                        \
    compose (series<C> f, series<C> g,             \
             HAS_CHARACTERISTIC dummy1, HAS_DIVISIBLE dummy2)
#define CHARACTERISTIC_COMPOSE(CHARACT_LT_1000)       \
    inline static series<C>                           \
    characteristic_compose (series<C> f, series<C> g, \
	                    CHARACTERISTIC_LT_1000 dummy)
template<class C> series<C>
compose (series<C> f, series<C> g) {
  template<class C> struct internal {    
    COMPOSE(bool_to_type<TRUE>,bool_to_type_any) {
      return characteristic_compose
	(f, g, bool_to_type<(characteristic<C>::value < 1000)>()); }
    CHARACTERISTIC_COMPOSE(bool_to_type<TRUE>) {
      return new compose3_series_rep<C> (f, g); }
    CARACTERISTIC_COMPOSE(bool_to_type<FALSE>) {
      return new compose2_series_rep<C> (f, g); }
    COMPOSE(bool_to_type<FALSE>,bool_to_type<TRUE>) {
      return new compose2_series_rep<C> (f, g); }
    COMPOSE(bool_to_type<FALSE>,bool_to_type<FALSE>) {
      return new compose1_series_rep<C> (f, g); }
  };
  if (!(g[0]==0)) fatal_error ("invalid composition", "compose");
  return internal::compose (f, g,
			    characteristic<C>::defined,
			    has_divisible<C>());
}
#undef COMPOSE
#undef CHARACTERISTIC_COMPOSE

#endif // defined TRUNCATED_COMPOSE_CC
