
/******************************************************************************
* MODULE     : truncated.gen.cc
* DESCRIPTION: truncated relaxed power series
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef TRUNCATED_CC
#define TRUNCATED_CC
#include "Univariate/relaxed.hh"
#include "Univariate/Zealous/zealous.cc"
#include "multiply.cc"
#include "compose.cc"
#include "series.cc"

/******************************************************************************
* general routines
******************************************************************************/

template<class C> C
series<C>::operator [] (int n) {
  if (n < rep->n) return rep->value[n];
  if (n >= N (rep->value)) rep->value= rep->value (0, 1 >? (n<<1));
  while (rep->n <= n) {
    rep->value[rep->n]= rep->next ();
    rep->n++;
  }
  // cout << ((tree) (*this)) << " [" << n << "] -> " << rep->value[n] << "\n";
  return rep->value[n];
}

template<class C> zealous<C>
series<C>::operator () (int start, int end) {
  if (end <= rep->n) return rep->value (start, end);
  if (end > N (rep->value)) rep->value= rep->value (0, end);
  while (rep->n < end) {
    rep->value[rep->n]= rep->next ();
    rep->n++;
  }
  return rep->value (start, end);
}

template<class C> series<C>::operator tree () {
  return (tree) (*rep);
}

/******************************************************************************
* implicit series
******************************************************************************/

template<class C> void
implicit_series_rep<C>::set_precision (int terms) {
  if (N(value)<terms) {
    value= value (0, terms);
    eq->set_precision (terms);
  }
}

template<class C> implicit_series<C>::implicit_series (int r) {
  rep= new implicit_series_rep<C> (r);
  rep->ref_count++;
}

template<class C> C&
implicit_series<C>::operator () (int n) {
  if (n>=rep->r) fatal_error ("Insufficient number of initial conditions",
			      "implicit_series<C>::operator []");
  return rep->value[n];
}

template<class C> C
implicit_series<C>::operator [] (int n) {
  return ((series<C>) (*this)) [n];
}

template<class C> zealous<C>
implicit_series<C>::operator () (int start, int end) {
  return ((series<C>) (*this)) (start, end);
}

template<class C> struct auxiliary_series_rep : public series_rep<C> {
  implicit_series<C> f;
  tree t;
  auxiliary_series_rep (implicit_series<C> f2, tree t2):
    series_rep<C> (RELAXED), f(f2), t (t2) {}
  ~auxiliary_series_rep () { f->eq= series<C>(); }
  C next () { return ((series<C>) f)[n]; }
  void set_precision (int terms) {
    if (N(value)<terms) {
      value= value (0, terms);
      f->set_precision (terms);
    }
  }
  operator tree () { return t; }
};

template<class C> series<C>
auxiliary_series<C> (implicit_series<C> f, tree t) {
  return new auxiliary_series_rep<C> (f, t);
}

/******************************************************************************
* constant and polynomial series
******************************************************************************/

template<class C> struct polynomial_series_rep : public series_rep<C> {
  polynomial_series_rep (zealous<C> f):
    series_rep<C> (POLYNOMIAL) { value= f; n= N(f); }
  C next () { return ZERO<C>(); }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms); }
  operator tree () { return tree ("polynomial"); }
};

template<class C> series<C>::series (C c, int deg) {
  zealous<C> f (deg+1);
  f[deg]= c;
  rep= new polynomial_series_rep<C> (f);
  rep->ref_count++;
}

template<class C> series<C>::series (zealous<C> f) {
  rep= new polynomial_series_rep<C> (f);
  rep->ref_count++;
}

/******************************************************************************
* series whose coefficients are given by a function
******************************************************************************/

template<class C> struct func_series_rep : public series_rep<C> {
  C (*coeffs) (int);
  func_series_rep (C (*coeffs2) (int)):
    series_rep<C> (RELAXED), coeffs (coeffs2) {}
  C next () { return coeffs(n); }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms); }
  operator tree () { return tree ("func"); }
};

template<class C> series<C>::series (C (*coeffs) (int)) {
  rep= new func_series_rep<C> (coeffs);
  rep->ref_count++;
}

/******************************************************************************
* shifting
******************************************************************************/

template<class C> struct shift_series_rep : public series_rep<C> {
  series<C> f;
  int sh;
  shift_series_rep (series<C> f2, int sh2):
    series_rep<C> (RELAXED), f(f2), sh(sh2) {}
  C next () {
    if ((n+sh)<0) return ZERO<C>();
    return f[n+sh];
  }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    if ((terms+sh)>0) f->set_precision (terms+sh);
  }
  operator tree () { return tuple (">>", f, tree (as_string (sh))); }
};

template<class C> series<C>
operator >> (series<C> f, int n) {
  return new shift_series_rep<C> (f, n);
}

template<class C> series<C>
operator << (series<C> f, int n) {
  return new shift_series_rep<C> (f, -n);
}

/******************************************************************************
* sums, differences and scalar products
******************************************************************************/

template<class C> struct sum_series_rep : public series_rep<C> {
  series<C> f, g;
  sum_series_rep<C> (series<C> f2, series<C> g2):
    series_rep<C> (RELAXED), f(f2), g(g2) {}
  C next () { return f[n]+ g[n]; }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    f->set_precision (terms);
    g->set_precision (terms);
  }
  operator tree () { return tuple ("+", f, g); }
};

template<class C> series<C>
operator + (series<C> f, series<C> g) {
  return new sum_series_rep<C> (f, g);
}

template<class C> struct diff_series_rep : public series_rep<C> {
  series<C> f, g;
  diff_series_rep (series<C> f2, series<C> g2):
    series_rep<C> (RELAXED), f(f2), g(g2) {}
  C next () { return f[n]- g[n]; }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    f->set_precision (terms);
    g->set_precision (terms);
  }
  operator tree () { return tuple ("-", f, g); }
};

template<class C> series<C>
operator - (series<C> f, series<C> g) {
  return new diff_series_rep<C> (f, g);
}

template<class C> struct scalar_mul_series_rep : public series_rep<C> {
  C c; series<C> f;
  scalar_mul_series_rep (C c2, series<C> f2):
    series_rep<C> (RELAXED), c (c2), f(f2) {}
  C next () { return c*f[n]; }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    f->set_precision (terms);
  }
  operator tree () { return tuple ("*", c, f); }
};

template<class C> series<C>
operator * (C c, series<C> f) {
  return new scalar_mul_series_rep<C> (c, f);
}

/******************************************************************************
* quotients
******************************************************************************/

template<class C> struct aux_inv_series_rep : public series_rep<C> {
  series<C> f;
  aux_inv_series_rep (series<C> f2):
    series_rep<C> (RELAXED), f(f2) {}
  C next () { return -f[n+1]/f[0]; }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    f->set_precision (terms+1);
  }
  operator tree () { return tuple ("aux-inv", f); }
};

template<class C> series<C>
operator / (series<C> f, series<C> g) {
  implicit_series<C> h (1);
  h(0) = C(1)/g[0];
  h->eq= ((new aux_inv_series_rep<C> (g)) * h) << 1;
  return f * auxiliary_series<C> (h, tuple ("/", f, g));
}

/******************************************************************************
* differential calculus
******************************************************************************/

template<class C> struct differentiated_series_rep : public series_rep<C> {
  series<C> f;
  differentiated_series_rep (series<C> f2):
    series_rep<C> (RELAXED), f(f2) {}
  C next () { return C(n+1)*f[n+1]; }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    f->set_precision (terms+1);
  }
  operator tree () { return tuple ("d", f); }
};

series<C>
d (series<C> f) {
  return new differentiated_series_rep<C> (f);
}

// if has_divisible<C>()

template<class C> struct scalar_div_series_rep : public series_rep<C> {
  series<C> f; C c;
  scalar_div_series_rep (series<C> f2, C c2):
    series_rep<C> (RELAXED), f(f2), c (c2) {}
  C next () { return f[n]/c; }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    f->set_precision (terms);
  }
  operator tree () { return tuple ("/", f, c); }
};

template<class C> series<C>
operator / (series<C> f, C c) {
  return new scalar_div_series_rep<C> (f, c);
}

template<class C> struct integrated_series_rep : public series_rep<C> {
  series<C> f;
  integrated_series_rep (series<C> f2):
    series_rep<C> (RELAXED), f(f2) {}
  C next () { if (n==0) return 0; return f[n-1]/C(n); }
  void set_precision (int terms) {
    if (N(value)<terms) value= value (0, terms);
    f->set_precision (terms-1);
  }
  operator tree () { return tuple ("I", f); }
};

template<class C> series<C>
I (series<C> f) {
  return new integrated_series_rep<C> (f);
}

template<class C> series<C>
exp (series<C> f) {
  implicit_series<C> g (1);
  if (has_special_functions<C>())
    g(0) = exp (f[0]);
  else
    g(0) = ONE<C>();
  g->eq= I(g*d(f));
  return auxiliary_series<C> (g, tuple ("exp", f));
}

template<class C> series<C>
log (series<C> f) {
  if (has_special_functions<C>())
    return series<C>(log (f[0]))+ I(d(f)/f);
  else
    return I(d(f)/f);
}

// end if has_divisible<C>()

/******************************************************************************
* output
******************************************************************************/

#define RELAXED_CODE(C) SERIES_CODE(C,10)

template<class C> void
series::set_print_precision (int n) {
  series<C>::print_precision= n;
}

template<class C> ostream&
operator << (ostream& out, series<C> f)
{ // this function is only for testing
  f->set_precision (series<C>::print_precision+10);
  return out << f(0, series<C>::print_precision);
}

#endif // defined TRUNCATED_CC
